#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <xc.h>

// linker: --rom=-FE0-FFF 

// CONFIG1 
#pragma config FOSC = INTOSC    // Oscillator Selection Bits (INTOSC oscillator: I/O function on CLKIN pin) 
#pragma config WDTE = OFF       // Watchdog Timer Enable (WDT disabled) 
#pragma config PWRTE = ON      // Power-up Timer Enable (PWRT disabled) 
#pragma config MCLRE = ON       // MCLR Pin Function Select (MCLR/VPP pin function is MCLR) 
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled) 
#pragma config BOREN = OFF      // Brown-out Reset Enable (Brown-out Reset disabled) 
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin) 

// CONFIG2 
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off) 
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset) 
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.) 
#pragma config LPBOR = OFF      // Low-Power Brown Out Reset (Low-Power BOR is disabled) 
#pragma config LVP = ON        // Low-Voltage Programming Enable (High-voltage on MCLR/VPP must be used for programming) 
#pragma config PPS1WAY = 0FF 

/*
#define rda_87_108MHz	0
#define rda_76_91MHz	1
#define rda_76_108MHz	2
#define rda_65_76MHz	3

#define rda_100kHz		0
#define rda_200kHz		1
#define rda_50kHz		2
#define rda_25kHz		3
*/

#define RS PORTCbits.RC5  //display text - command  
#define EN PORTAbits.RA1  //display enable 

#define CMD 0  //command
#define DAT 1    //data



static const char *txt[] = {
    "RDA5807M OK",
    "FM Radio   ",
    "   *    ",
    "  ---   ",
    "            ",
    "////////",
    "Writing!    "
};

static const char *stn[] = {
    "COUNTRY ",
    "R-DVOJKA",
    "R-ZURNAL",
    "HITRADIO",
    "PROGLAS,",
    "CONTACT ",
    "R-VLTAVA",
    "EVROPA 2",
    "R-SEVER ",
    "RADIO F1",
    " SIGNAL ",
    " IMPULS ",
    "R-REGION",
    "R-PLUS  ",
    "  BEAT  ",
    "  KISS  ",
    " BONTON ",
    "CLASSIC "
};


unsigned int freq = 0;
char i = 0, j = 0, ir;
char *buf;
char offset, ofso = 0, c1, c2, hasn = 0, prob = 0;

unsigned int cnt = 0x1000, rds = 0;
char len, slen, btn = 0, end = 0, menu = 0, btno = 0x80;
unsigned int chan;

unsigned int Band[5] = {870, 760, 760, 650, 500}; //begin freq.
char Space[4] = {2, 1, 4, 8};
char wbuf[16];
char rbuf[16];
char stname[16];
char str[24];

//Band and spacing 
unsigned int startf = 892; //default freq.
char band = 0; //0 = 87-108MHz, 1 = 76-91MHz, 2 = 76-108MHz, 3 = 50/65-76MHz
char spacing = 0; //0=100kHz, 1=200kHz, 2=50kHz 3=25kHz

/*----------------------------------------------------------*/
struct REG2h { //wbuf 0
    char SEEK : 1; // 1   
    char SEEKUP : 1; // 2
    char RCKL_DIM : 1; // 4
    char RCLK : 1; // 8
    char BASS : 1; // 0x10
    char MONO : 1; // 0x20 
    char DMUTE : 1; // 0x40
    char DHIZ : 1; // 0x80
} *Reg2h = (struct REG2h*) &wbuf[0];

struct REG2 { //wbuf 1
    char ENABLE : 1; //1     
    char SOFT_RESET : 1; //2
    char NEW_METHOD : 1; //4
    char RDS_EN : 1; //8
    char CLK_MODE : 3; //0x70
    char SKMODE : 1; //0x80
} *Reg2 = (struct REG2*) &wbuf[1];

/*
struct REG3h {
    char CHAN : 8;      //wbuf 2
};
struct REG3h *Reg3h = (struct REG3h*) &wbuf[2];
 */
struct REG3 { //wbuf 3
    char SPACE : 2;
    char BAND : 2;
    char TUNE : 1;
    char DIR_MODE : 1;
    char CHANL : 2;
} *Reg3 = (struct REG3*) &wbuf[3];

struct REG4h { //wbuf 4
    char AFCD : 1;
    char SOTFMUTE : 1;
    char : 1;
    char DE : 1;
    char : 2;
    char : 2;
} *Reg4h = (struct REG4h*) &wbuf[4];

/*
struct REG4 {
    char : 8;           //wbuf 5
};
struct REG4 *Reg4 = (struct REG4*) &wbuf[5];
 */
struct REG5h { //wbuf 6
    char SEEKTH : 4;
    char : 3;
    char INT_MODE : 1;
} *Reg5h = (struct REG5h*) &wbuf[6];

struct REG5 { //wbuf 7
    char VOLUME : 4;
    char LNA : 4;
} *Reg5 = (struct REG5*) &wbuf[7];

/*
struct REG6h {
    char : 8;           //wbuf 8 
};
struct REG6h *Reg6h = (struct REG6h*) &wbuf[8];

struct REG6 {
    char : 8;           //wbuf 9 
};
struct REG6 *Reg6 = (struct REG6*) &wbuf[9];
 */
struct REG7h { //wbuf 10
    char : 1;
    char M65_50 : 1;
    char TH_SOFTBLEND : 5;
    char : 1;
} *Reg7h = (struct REG7h*) &wbuf[10];

struct REG7 { //wbuf 11
    char FREQ_MODE : 1;
    char SOFTBLEND : 1;
    char SEEK_TH_OLD : 6;
} *Reg7 = (struct REG7*) &wbuf[11];

struct REGAh { //rbuf 0
    char RCHANL : 2;
    char ST : 1;
    char BLK_E : 1;
    char RDSS : 1;
    char SF : 1;
    char STC : 1;
    char RDSR : 1;
} *RegAh = (struct REGAh*) &rbuf[0];

/*
struct REGA {           //rbuf 1  RCHAN 
    char RCHAN : 8;
};
struct REGA *RegA = (struct REGA*) &rbuf[1];
 */
struct REGBh { //rbuf 2
    char FMTRUE : 1;
    char RSSI : 7;
} *RegBh = (struct REGBh*) &rbuf[2];

struct REGB { //rbuf 3
    char BLERB : 2;
    char BLERA : 2;
    char ABCD_E : 1;
    char : 2;
    char FM_READY : 1;
} *RegB = (struct REGB*) &rbuf[3];

/*---------------------------------------------*/

void Delay(int tms) { //ms
    while (tms) {
        _delay(4000);
        tms--;
    }
}

void Sdelay(char dl) {
    while (dl) {
        _delay(32); //8us
        dl--;
    }
}

void LCD_Start(char ch) {
    PORTC = (char) (ch & 0x0F);
    EN = 0;
    EN = 1;
    if (ch == 3) Sdelay(200); //1,5 ms
    else Sdelay(4); //32us
}

void LCD_Chr(char dchr, char dc) {

    RS = dc; //RS = 1(data)
    PORTC &= 0x30;
    PORTC |= (char) ((dchr & 0xF0) >> 4);
    EN = 1;
    EN = 0;
    PORTC &= 0x30;
    PORTC |= (char) (dchr & 0x0F);
    EN = 1;
    EN = 0;

    if (dc == CMD && (dchr < 8)) Sdelay(200); // __delay_ms(2);
    else Sdelay(4); //40us
    EN = 1;

}

void LCD_Init(void) {
    Delay(40);

    EN = 1;
    RS = 0;
    LCD_Start(3); //wait > 4,1ms
    LCD_Start(3);
    LCD_Start(3);
    LCD_Start(2);

    LCD_Chr(0x28, CMD); //2 rows
    LCD_Chr(0x10, CMD); //cursor left
    LCD_Chr(0x0C, CMD); //cursor off, disp on   (0xE0 cursor on)
    LCD_Chr(0x06, CMD); //increment
    LCD_Chr(0x01, CMD); //clear LCD

}

void LCD_Strx(char *str) {
    while (*str != 0)
        LCD_Chr(*str++, DAT);
}

char butn() {
    char but;
    TRISC = 0x3F;
    Sdelay(5); //40us
    but = (char) ((~PORTC) & 0x3F);
    TRISC = 0;
    return but;
}

void write_rda() {
    buf = wbuf;
    SEN = 1; //start I2C
    while (SEN);
    SSP1IF = 0;
    SSPBUF = 0x20; //adr << 1; adr rrd = 0x10 
    while (!SSP1IF);
    for (ir = 0; ir < 12; ir++) {
        SSP1IF = 0;
        SSPBUF = *buf;
        buf++;
        while (!SSP1IF);
    }

    PEN = 1; //stop I2C
    while (PEN);
}

void read_rda() {
    buf = rbuf;
    SEN = 1; //start I2C
    while (SEN);

    SSP1IF = 0;
    SSPBUF = 0x21; //((adr << 1) | 1);
    while (!SSP1IF);

    for (ir = 0; ir < 12; ir++) {
        RCEN = 1; //enable read  ;
        if (ir < 12 - 1)
            ACKDT = 0; //next byte 
        else ACKDT = 1; // no next byte  
        while (!BF);
        *buf = SSPBUF;
        ACKEN = 1; // sendind ACK or NACK
        while (ACKEN);
        buf++;
    }
    RCEN = 0;

    PEN = 1; //stop I2C 
    while (PEN);
}

void write_setting() {
    // REG 02
    Reg2->CLK_MODE = 0;
    Reg2->ENABLE = 1;
    Reg2->NEW_METHOD = 1;
    Reg2->RDS_EN = 1;
    Reg2->SOFT_RESET = 0;
    Reg2->SKMODE = 1;
    Reg2h->DMUTE = 1;
    Reg2h->SEEKUP = 1;
    Reg2h->DHIZ = 1;
    Reg2h->MONO = 0;

    // REG 03 - auto tune, 87-108 band, 0.1 spacing   
    chan = ((startf - Band[band]) * Space[spacing]) / 2; //default freq.
    wbuf[2] = chan >> 2;
    Reg3->CHANL = chan & 3;
    Reg3->DIR_MODE = 0;
    Reg3->BAND = (char) (band == 4 ? 3 : band);
    Reg3->SPACE = spacing;
    Reg3->TUNE = 1;

    // REG 04 - audio 50us, no soft mute, no disable AFC   
    Reg4h->DE = 0;
    Reg4h->SOTFMUTE = 0;
    Reg4h->AFCD = 0;

    wbuf[5] = 0;

    // REG 05 - max volume  LNA = 8, gain?
    Reg5->VOLUME = 15;
    Reg5->LNA = 8;
    Reg5h->INT_MODE = 1;
    Reg5h->SEEKTH = 8; // 32 dB S/N

    // REG 06 - reserved
    wbuf[8] = 0;
    wbuf[9] = 0;

    // REG 07
    Reg7->FREQ_MODE = 0;
    Reg7->SOFTBLEND = 1;
    Reg7->SEEK_TH_OLD = 0;
    Reg7h->TH_SOFTBLEND = 0x0F;
    Reg7h->M65_50 = (char) (band == 4 ? 0 : 1);

    write_rda();
}

void MemRead() {
    char adrl = 0xE0;
    PMADRH = 0x0F; //Store MSB of address
    for (i = 0; i < 12; i++) {
        PMADRL = adrl; //Store LSB of address    
        PMCON1bits.CFGS = 0; //Do not select Configuration Space
        PMCON1bits.RD = 1; //Initiate read
        asm("NOP"); //Ignored (Figure 10-2)
        asm("NOP"); //Ignored (Figure 10-2)
        rbuf[i] = PMDATL;
        adrl++;
    }
}

void Unlock() {
    PMCON2 = 0x55; // Write 55h
    PMCON2 = 0xAA; //Write AAh
    PMCON1bits.WR = 1; // Set WR bit to begin write
    asm("NOP");
    asm("NOP");
}

void MemWrite(char *dat) {
    char adrl = 0xE0;
    GIE = 0; //Disable ints so required sequences will execute properly
    PMADRH = 0xF; //Store MSB of address
    for (i = 0; i < 12; i++) {
        PMADRL = adrl;
        PMCON1bits.CFGS = 0; //Do not select Configuration Space
        PMCON1bits.FREE = 0; //not Specify an erase operation
        PMCON1bits.LWLO = 1; // Only Load Write Latches
        PMCON1bits.WREN = 1; //Initiate Enable writes
        PMDATL = *dat;
        PMDATH = 0;
        dat++;
        adrl++;
        Unlock();
    }

    PMCON1bits.LWLO = 0; //No more loading latches - Actually start Flash program
    Unlock();
    PMCON1bits.WREN = 0; //Disable writes
}

void MemEras() {
    GIE = 0; //Disable ints so required sequences will execute properly
    PMADRL = 0xE0;
    PMADRH = 0xF;
    PMCON1bits.CFGS = 0; //Not configuration space
    PMCON1bits.FREE = 1; //Specify an erase operation
    PMCON1bits.WREN = 1; //Enable writes
    Unlock();
    PMCON1bits.WREN = 0; //Disable writes
}

/*------------------ MAIN ---------------------*/

void main() {

    OSCCON = 0x78; //16 MHz
    OPTION_REGbits.nWPUEN = 0;
    PORTA = 0x0;
    PORTC = 0;
    LATA = 0;
    LATC = 0;
    ANSELA = 0;
    ANSELC = 0;
    TRISA = 0x38;
    TRISC = 0; // Port C 

    /*------- PPS --------*/
    GIE = 0;
    PPSLOCK = 0x55;
    PPSLOCK = 0xAA;
    PPSLOCKED = 0; //enable PPS
    /* I2C SCL - RA4, SDA - RA5*/
    RA4PPSbits.RA4PPS = 0x10; //RA4->MSSP:SCL;
    SSPDATPPSbits.SSPDATPPS = 0x05; //RA5->MSSP:SDA;
    SSPCLKPPSbits.SSPCLKPPS = 0x04; //RA4->MSSP:SCL;
    RA5PPSbits.RA5PPS = 0x11; //RA5->MSSP:SDA;   
    RA0PPSbits.RA0PPS = 0x0E; //RA0->PWM3:PWM3OUT;
    PPSLOCKED = 1; //dis. PPS
    /*-----------------------------*/
    Delay(800);
    LCD_Init();

    /*------------- Open I2C -----------*/
    SSPSTAT &= 0x3F; // power on state
    SSPCON = 0x08; //  select serial mode                
    SMP = 1; // slew rate  0 = high speed 400kHz, 1 = standard 100kHz or 1MHz
    SSPADD = 0x27; //0x27 100kHz  0x09 pro 400kHz   
    SSPEN = 1; // enable synchronous serial port

    /*---------- PWM Display ---------------*/
    TMR2IE = 0;
    PWM3CON = 0xF;
    PR2 = 249;
    PWM3DCL = 0;
    PWM3DCH = 0x80;
    T2CON = 0x4E; //x6;
    PWM3EN = 1;

    Delay(100);
    read_rda(); //read chip ID 

    if (rbuf[6] == 0x58 && rbuf[7] == 04 && rbuf[10] == 0x58 && rbuf[11] == 04)
        LCD_Strx((char*) txt[0]); //display chip ID on powerup

    slen = (sizeof (stn)) / (sizeof (stn[0])); //number of stations
    Delay(1000);

    MemRead();
    if (rbuf[0] == 0xFF)
        write_setting(); //implicit setting
    else {
        for (i = 0; i < 12; i++) {
            wbuf[i] = rbuf[i];
        }
        Reg3->TUNE = 1;
        write_rda(); // setting from FLASH 0xFE0
    }

    LCD_Chr(0x80, CMD);
    LCD_Strx((char *) txt[1]); //display FM RADIO
    Delay(1000);

    while (1) { //main loop

        btn = butn(); // read buttons
        if (btn == 0)btno = 0x80;
        if (btn && btn != btno) {
            if (btn == 0x20) {
                if (menu > 2) {
                    menu = 0;
                    end = 0;
                    LCD_Chr(0xC0, CMD); //kursor
                    LCD_Strx((char *) txt[4]);
                    chan = ((freq / 10 - Band[band]) * Space[spacing]) / 2;
                    chan &= 0x3FF;
                    wbuf[2] = chan >> 2; //  Reg3h->CHAN = chan >> 2;
                    Reg3->CHANL = chan & 3;
                    Reg3->TUNE = 1;
                } else menu++;
            }//btn 0x20

            if (menu == 0) {
                switch (btn) {
                    case 0x10:
                        Reg2h->DMUTE ^= 1;
                        break;
                    case 0x08: if (Reg5->VOLUME > 0) Reg5->VOLUME--;
                        break;
                    case 0x04: if (Reg5->VOLUME < 15) Reg5->VOLUME++;
                        break;
                    case 0x02: Reg2h->SEEKUP = 0;
                        Reg2h->SEEK = 1;
                        break;
                    case 0x01: Reg2h->SEEKUP = 1;
                        Reg2h->SEEK = 1;
                        break;
                }//switch
            }// menu 0
            else if (menu == 1) {
                switch (btn) {
                    case 0x10: chan = 0; //band begin
                        break;
                    case 0x08: chan -= 10;
                        break;
                    case 0x04: chan += 10;
                        break;
                    case 0x02: chan -= 1;
                        break;
                    case 0x01: chan += 1;
                        break;
                } //switch

                chan &= 0x3FF;
                wbuf[2] = chan >> 2; //  Reg3h->CHAN = chan >> 2;
                Reg3->CHANL = chan & 3;
                Reg3->TUNE = 1;

            }//menu 1
            else if (menu == 2) {
                switch (btn) {
                    case 0x10: Reg2->RDS_EN ^= 1;
                        break;
                    case 0x08: Reg2h->BASS ^= 1;
                        break;
                    case 0x04: wbuf[0] ^= 0x20; //Reg2h->MONO ^= 1;                                                
                        break;
                    case 0x02: Reg5h->SEEKTH++;
                        break;
                    case 0x01: Reg4h->DE ^= 1;
                        break;
                } //switch  
            }//menu 2
            else if (menu == 3) {
                switch (btn) {
                    case 0x10: Reg2->SOFT_RESET = 1;
                        MemEras();
                        break;
                    case 0x08: band++;
                        if (band > 4)band = 0;
                        Reg3->BAND = (char) (band == 4 ? 3:band);
                        Reg7h->M65_50 = (char) (band == 4 ? 0 : 1);
                        break;
                    case 0x04: spacing++;
                        if (spacing > 3)spacing = 0;
                        Reg3->SPACE = spacing;
                        break;
                    case 0x02: Reg4h->AFCD ^= 1;
                        break;
                    case 0x01: MemEras();
                        MemWrite(wbuf);
                        LCD_Chr(0xC0, CMD);
                        LCD_Strx((char *) txt[6]); //display FM RADIO
                        Delay(1000);
                        break;
                }//switch  
            }//menu 3

            write_rda();
            btno = btn;
        }// if btn

        if (Reg2->SOFT_RESET) asm("RESET"); //Restar incl. chip test

        if (end == 1)Delay(100);

        /*-------- Seek or Tune ------*/
        read_rda(); //read chip

        if (Reg2h->SEEK || Reg3->TUNE) {
            LCD_Chr(0xC0, CMD); //cursor
            LCD_Strx((char *) txt[4]); //clear display stname

            /*----- display freq -----*/

            chan = (unsigned) ((rbuf[0] << 8 | rbuf[1]) &0x3FF);
            wbuf[2] = chan >> 2;
            Reg3->CHANL = chan & 3;
            freq = (Band[band]*10) + (chan * 20) / Space[spacing];
            itoa(str, freq, 10);
            len = strlen((const char*) str);
            str[len] = str[(char) (len - 1)];
            str[(char) (len - 1)] = str[(char) (len - 2)];
            str[(char) (len - 2)] = '.';
            str[(char) (len + 1)] = 0;
            strcat(str, " MHz ");
            LCD_Chr(0x80, CMD); //kursor
            LCD_Strx(str);

            Delay(35);
            if (RegAh->STC) { //station found
                Reg2h->SEEK = 0;
                Reg3->TUNE = 0;
                hasn = 0;
                end = 0;
                cnt = 0x1000; //search counter
                rds = 0;
                strcpy(stname, txt[5]);
                write_rda();
                //   Delay(35);

            } //if STC
        }//if SEEK || TUNE

        /*---------- Stereo/mono, Smeter --------------*/
        itoa(&str[3], (RegBh->RSSI + 5) / 10, 10);

        if (str[4] == 0) {
            str[4] = str[3];
            str[3] = '=';
            str[5] = 0;
        }
        str[0] = RegAh->ST ? 's' : 'm';
        str[1] = ' ';
        str[2] = 'S';
        LCD_Chr(0x8B, CMD); //set cursor
        LCD_Strx(str); //display S

        /*------- Display settings --------*/
        if (menu == 0) {
            itoa(&str[2], Reg5->VOLUME, 10);
            strcat(str, "  ");
            str[0] = ' ';
            str[1] = 'V';
        } else if (menu == 1)strcpy(str, "MAN ");
        else if (menu == 2) {
            strcpy(str, "SET1");
            strcpy(stname, "r b S T  D ");
            if (Reg2->RDS_EN)stname[0] = 'R';
            if (Reg2h->BASS)stname[2] = 'B';
            if (Reg2h->MONO)stname[4] = 'M';
            stname[7] = (char) (Reg5h->SEEKTH + (Reg5h->SEEKTH > 9 ? 0x37 : 0x30));
            stname[10] = Reg4h->DE ? '7' : '5';

            LCD_Chr(0xC0, CMD);
            LCD_Strx(stname);
        } else if (menu == 3) {
            strcpy(str, "SET2");
            strcpy(stname, "Q B  S  A W ");
            stname[3] = (char) (band + 0x30);
            stname[6] = (char) (spacing + 0x30); // spacing 0 1 2 3 
            if (Reg4h->AFCD)stname[8] = 'a'; // AFCD
            LCD_Chr(0xC0, CMD);
            LCD_Strx(stname);
        }

        if (menu == 0 && Reg2->RDS_EN == 0) {
            LCD_Chr(0xC0, CMD); //set cursor
            LCD_Strx((char *) txt[3]); //display "---"  
        }

        LCD_Chr(0xCC, CMD); //set cursor
        if (!Reg2h->DMUTE)strcpy(str, "MUTE");
        LCD_Strx(str); //display VOL or MUTE   

        /*-------- RDS Searching ------*/
        if (menu < 2 && end == 0 && Reg2->RDS_EN) {
            read_rda();
            if (/*RegAh->RDSS && */ RegAh->RDSR && RegB->BLERA == 0 && RegB->BLERB == 0) {
                if ((rbuf[6] & 0xF8) == 0 /*|| (rbuf[6] & 0xF8) == 0x08 */) {
                    offset = (char) ((rbuf[7] & 3) << 1); //character pointer
                    c1 = (char) (rbuf[10]); //  Block D 1st char                 
                    c2 = (char) (rbuf[11]); //  Block D 2nd char

                    if (offset != ofso) { //new chars
                        stname[offset] = c1;
                        stname[(char) (offset + 1)] = c2;
                        stname[8] = 0;

                        for (i = 0; i < slen; i++) {
                            prob = 0;
                            for (j = 0; j < 8; j++) {
                                if (stname[j] == stn[i][j]) prob++;
                            }
                            if (prob > 5) {
                                LCD_Chr(0xC0, CMD);
                                LCD_Strx((char *) stn[i]);
                                hasn = 1;
                                end = 1;
                            }//if prob
                        }//for i					
                    }//if offset!=ofso
                    ofso = offset;
                }//if Block
                rds++;
            }//if RDSS
            if (rds) {
                cnt += rds / 2, rds = 0;
            }
            cnt--;
            if (cnt == 1)end = 1; //end of searching 

            /*------- Blinking Star ------*/
            if (hasn == 0) {
                if ((cnt & 0x007F) == 0x0040) {
                    LCD_Chr(0xC0, CMD);
                    strcpy(str, txt[2]);
                    if (cnt & 0x80)str[3] = ' ';
                    LCD_Strx(str);
                }
            }
        }//if RDSEN  

    }//while 1

    SSPEN = 0; //close I2C
}// main


